from abc import ABCMeta, abstractmethod
from typing import List

from win2xcur.cursor import CursorFrame


class BaseParser(metaclass=ABCMeta):
    blob: bytes
    frames: List[CursorFrame]

    @abstractmethod
    def __init__(self, blob: bytes) -> None:
        self.blob = blob

    @classmethod
    @abstractmethod
    def can_parse(cls, blob: bytes) -> bool:
        raise NotImplementedError()



from typing import List, Type

from win2xcur.parser.ani import ANIParser
from win2xcur.parser.base import BaseParser
from win2xcur.parser.cur import CURParser
from win2xcur.parser.xcursor import XCursorParser

__all__ = ['ANIParser', 'CURParser', 'XCursorParser', 'PARSERS', 'open_blob']

PARSERS: List[Type[BaseParser]] = [CURParser, ANIParser, XCursorParser]


def open_blob(blob: bytes) -> BaseParser:
    for parser in PARSERS:
        if parser.can_parse(blob):
            return parser(blob)
    raise ValueError('Unsupported file format')



from typing import List

from wand.color import Color
from wand.image import BaseImage, COMPOSITE_OPERATORS, Image

from win2xcur.cursor import CursorFrame

if 'copy_opacity' in COMPOSITE_OPERATORS:
    COPY_ALPHA = 'copy_opacity'  # ImageMagick 6 name
    NEEDS_NEGATE = False
else:
    COPY_ALPHA = 'copy_alpha'  # ImageMagick 7 name
    NEEDS_NEGATE = True


def apply_to_image(image: BaseImage, *, color: str, radius: float, sigma: float, xoffset: float,
                   yoffset: float) -> Image:
    xoffset = round(xoffset * image.width)
    yoffset = round(yoffset * image.height)
    new_width = image.width + 3 * xoffset
    new_height = image.height + 3 * yoffset

    if NEEDS_NEGATE:
        channel = image.channel_images['opacity'].clone()
        channel.negate()
    else:
        channel = image.channel_images['opacity']

    opacity = Image(width=new_width, height=new_height, pseudo='xc:white')
    opacity.composite(channel, left=xoffset, top=yoffset)
    opacity.gaussian_blur(radius * image.width, sigma * image.width)
    opacity.negate()
    opacity.modulate(50)

    shadow = Image(width=new_width, height=new_height, pseudo='xc:' + color)
    shadow.composite(opacity, operator=COPY_ALPHA)

    result = Image(width=new_width, height=new_height, pseudo='xc:transparent')
    result.composite(shadow)
    result.composite(image)

    trimmed = result.clone()
    trimmed.trim(color=Color('transparent'))
    result.crop(width=max(image.width, trimmed.width), height=max(image.height, trimmed.height))
    return result


def apply_to_frames(frames: List[CursorFrame], *, color: str, radius: float,
                    sigma: float, xoffset: float, yoffset: float) -> None:
    for frame in frames:
        for cursor in frame:
            cursor.image = apply_to_image(cursor.image, color=color, radius=radius,
                                          sigma=sigma, xoffset=xoffset, yoffset=yoffset)



import struct
from copy import copy
from typing import Any, Iterable, List, Tuple

from win2xcur.cursor import CursorFrame
from win2xcur.parser.base import BaseParser
from win2xcur.parser.cur import CURParser


class ANIParser(BaseParser):
    SIGNATURE = b'RIFF'
    ANI_TYPE = b'ACON'
    HEADER_CHUNK = b'anih'
    LIST_CHUNK = b'LIST'
    SEQ_CHUNK = b'seq '
    RATE_CHUNK = b'rate'
    FRAME_TYPE = b'fram'
    ICON_CHUNK = b'icon'
    RIFF_HEADER = struct.Struct('<4sI4s')
    CHUNK_HEADER = struct.Struct('<4sI')
    ANIH_HEADER = struct.Struct('<IIIIIIIII')
    UNSIGNED = struct.Struct('<I')
    SEQUENCE_FLAG = 0x2
    ICON_FLAG = 0x1

    @classmethod
    def can_parse(cls, blob: bytes) -> bool:
        signature: bytes
        size: int
        subtype: bytes
        try:
            signature, size, subtype = cls.RIFF_HEADER.unpack(blob[:cls.RIFF_HEADER.size])
        except struct.error:
            return False
        return signature == cls.SIGNATURE and subtype == cls.ANI_TYPE

    def __init__(self, blob: bytes) -> None:
        super().__init__(blob)
        if not self.can_parse(blob):
            raise ValueError('Not a .ani file')
        self.frames = self._parse(self.RIFF_HEADER.size)

    def _unpack(self, struct_cls: struct.Struct, offset: int) -> Tuple[Any, ...]:
        return struct_cls.unpack(self.blob[offset:offset + struct_cls.size])

    def _read_chunk(self, offset: int, expected: Iterable[bytes]) -> Tuple[bytes, int, int]:
        found = []
        while True:
            name, size = self._unpack(self.CHUNK_HEADER, offset)
            offset += self.CHUNK_HEADER.size
            if name in expected:
                break
            found += [name]
            offset += size
            if offset >= len(self.blob):
                raise ValueError(f'Expected chunk {expected!r}, found {found!r}')
        return name, size, offset

    def _parse(self, offset: int) -> List[CursorFrame]:
        _, size, offset = self._read_chunk(offset, expected=[self.HEADER_CHUNK])

        if size != self.ANIH_HEADER.size:
            raise ValueError(f'Unexpected anih header size {size}, expected {self.ANIH_HEADER.size}')

        size, frame_count, step_count, width, height, bit_count, planes, display_rate, flags = self.ANIH_HEADER.unpack(
            self.blob[offset:offset + self.ANIH_HEADER.size])

        if size != self.ANIH_HEADER.size:
            raise ValueError(f'Unexpected size in anih header {size}, expected {self.ANIH_HEADER.size}')

        if not flags & self.ICON_FLAG:
            raise NotImplementedError('Raw BMP images not supported.')

        offset += self.ANIH_HEADER.size

        frames = []
        order = list(range(frame_count))
        delays = [display_rate for _ in range(step_count)]

        while offset < len(self.blob):
            name, size, offset = self._read_chunk(offset, expected=[self.LIST_CHUNK, self.SEQ_CHUNK, self.RATE_CHUNK])
            if name == self.LIST_CHUNK:
                list_end = offset + size
                if self.blob[offset:offset + 4] != self.FRAME_TYPE:
                    raise ValueError(
                        f'Unexpected RIFF list type: {self.blob[offset:offset + 4]!r}, expected {self.FRAME_TYPE!r}')
                offset += 4

                for i in range(frame_count):
                    _, size, offset = self._read_chunk(offset, expected=[self.ICON_CHUNK])
                    frames.append(CURParser(self.blob[offset:offset + size]).frames[0])
                    offset += size
                    if offset & 1:
                        offset += 1

                if offset != list_end:
                    raise ValueError(f'Wrong RIFF list size: {offset}, expected {list_end}')
            elif name == self.SEQ_CHUNK:
                order = [i for i, in self.UNSIGNED.iter_unpack(self.blob[offset:offset + size])]
                if len(order) != step_count:
                    raise ValueError(f'Wrong animation sequence size: {len(order)}, expected {step_count}')
                offset += size
            elif name == self.RATE_CHUNK:
                delays = [i for i, in self.UNSIGNED.iter_unpack(self.blob[offset:offset + size])]
                if len(delays) != step_count:
                    raise ValueError(f'Wrong animation rate size: {len(delays)}, expected {step_count}')
                offset += size

        if len(order) != step_count:
            raise ValueError('Required chunk "seq " not found.')

        sequence = [copy(frames[i]) for i in order]
        for frame, delay in zip(sequence, delays):
            frame.delay = delay / 60

        return sequence



from itertools import chain
from operator import itemgetter
from typing import List

from win2xcur.cursor import CursorFrame
from win2xcur.parser import XCursorParser
from win2xcur.utils import premultiply_alpha


def to_x11(frames: List[CursorFrame]) -> bytes:
    chunks = []

    for frame in frames:
        for cursor in frame:
            hx, hy = cursor.hotspot
            header = XCursorParser.IMAGE_HEADER.pack(
                XCursorParser.IMAGE_HEADER.size,
                XCursorParser.CHUNK_IMAGE,
                cursor.nominal,
                1,
                cursor.image.width,
                cursor.image.height,
                hx,
                hy,
                int(frame.delay * 1000),
            )
            chunks.append((
                XCursorParser.CHUNK_IMAGE,
                cursor.nominal,
                header + premultiply_alpha(bytes(cursor.image.export_pixels(channel_map='BGRA')))
            ))

    header = XCursorParser.FILE_HEADER.pack(
        XCursorParser.MAGIC,
        XCursorParser.FILE_HEADER.size,
        XCursorParser.VERSION,
        len(chunks),
    )

    offset = XCursorParser.FILE_HEADER.size + len(chunks) * XCursorParser.TOC_CHUNK.size
    toc = []
    for chunk_type, chunk_subtype, chunk in chunks:
        toc.append(XCursorParser.TOC_CHUNK.pack(
            chunk_type,
            chunk_subtype,
            offset,
        ))
        offset += len(chunk)

    return b''.join(chain([header], toc, map(itemgetter(2), chunks)))



import struct
from collections import defaultdict
from typing import Any, Dict, List, Tuple, cast

from wand.image import Image

from win2xcur.cursor import CursorFrame, CursorImage
from win2xcur.parser.base import BaseParser


class XCursorParser(BaseParser):
    MAGIC = b'Xcur'
    VERSION = 0x1_0000
    FILE_HEADER = struct.Struct('<4sIII')
    TOC_CHUNK = struct.Struct('<III')
    CHUNK_IMAGE = 0xFFFD0002
    IMAGE_HEADER = struct.Struct('<IIIIIIIII')

    @classmethod
    def can_parse(cls, blob: bytes) -> bool:
        return blob[:len(cls.MAGIC)] == cls.MAGIC

    def __init__(self, blob: bytes) -> None:
        super().__init__(blob)
        self.frames = self._parse()

    def _unpack(self, struct_cls: struct.Struct, offset: int) -> Tuple[Any, ...]:
        return struct_cls.unpack(self.blob[offset:offset + struct_cls.size])

    def _parse(self) -> List[CursorFrame]:
        magic, header_size, version, toc_size = self._unpack(self.FILE_HEADER, 0)
        assert magic == self.MAGIC

        if version != self.VERSION:
            raise ValueError(f'Unsupported Xcursor version 0x{version:08x}')

        offset = self.FILE_HEADER.size
        chunks: List[Tuple[int, int, int]] = []
        for i in range(toc_size):
            chunk_type, chunk_subtype, position = self._unpack(self.TOC_CHUNK, offset)
            chunks.append((chunk_type, chunk_subtype, position))
            offset += self.TOC_CHUNK.size

        images_by_size: Dict[int, List[Tuple[CursorImage, int]]] = defaultdict(list)

        for chunk_type, chunk_subtype, position in chunks:
            if chunk_type != self.CHUNK_IMAGE:
                continue

            size, actual_type, nominal_size, version, width, height, x_offset, y_offset, delay = \
                self._unpack(self.IMAGE_HEADER, position)
            delay /= 1000

            if size != self.IMAGE_HEADER.size:
                raise ValueError(f'Unexpected size: {size}, expected {self.IMAGE_HEADER.size}')

            if actual_type != chunk_type:
                raise ValueError(f'Unexpected chunk type: {actual_type}, expected {chunk_type}')

            if nominal_size != chunk_subtype:
                raise ValueError(f'Unexpected nominal size: {nominal_size}, expected {chunk_subtype}')

            if width > 0x7FFF:
                raise ValueError(f'Image width too large: {width}')

            if height > 0x7FFF:
                raise ValueError(f'Image height too large: {height}')

            if x_offset > width:
                raise ValueError(f'Hotspot x-coordinate too large: {x_offset}')

            if y_offset > height:
                raise ValueError(f'Hotspot x-coordinate too large: {y_offset}')

            image_start = position + self.IMAGE_HEADER.size
            image_size = width * height * 4
            blob = self.blob[image_start:image_start + image_size]
            if len(blob) != image_size:
                raise ValueError(f'Invalid image at {image_start}: expected {image_size} bytes, got {len(blob)} bytes')

            image = Image(width=width, height=height)
            image.import_pixels(channel_map='BGRA', data=blob)
            images_by_size[nominal_size].append(
                (CursorImage(image.sequence[0], (x_offset, y_offset), nominal_size), delay)
            )

        if len(set(map(len, images_by_size.values()))) != 1:
            raise ValueError('win2xcur does not support animations where each size has different number of frames')

        result = []
        for sequence in cast(Any, zip(*images_by_size.values())):
            images: Tuple[CursorImage, ...]
            delays: Tuple[int, ...]
            images, delays = cast(Any, zip(*sequence))

            if len(set(delays)) != 1:
                raise ValueError('win2xcur does not support animations where each size has a different frame delay')

            result.append(CursorFrame(list(images), delays[0]))

        return result



from typing import Iterator, List, Tuple

from wand.sequence import SingleImage


class CursorImage:
    image: SingleImage
    hotspot: Tuple[int, int]
    nominal: int

    def __init__(self, image: SingleImage, hotspot: Tuple[int, int], nominal: int) -> None:
        self.image = image
        self.hotspot = hotspot
        self.nominal = nominal

    def __repr__(self) -> str:
        return f'CursorImage(image={self.image!r}, hotspot={self.hotspot!r}, nominal={self.nominal!r})'


class CursorFrame:
    images: List[CursorImage]
    delay: int

    def __init__(self, images: List[CursorImage], delay: int = 0) -> None:
        self.images = images
        self.delay = delay

    def __getitem__(self, item: int) -> CursorImage:
        return self.images[item]

    def __len__(self) -> int:
        return len(self.images)

    def __iter__(self) -> Iterator[CursorImage]:
        return iter(self.images)

    def __repr__(self) -> str:
        return f'CursorFrame(images={self.images!r}, delay={self.delay!r})'



from io import BytesIO
from itertools import chain
from typing import List, Tuple

from win2xcur.cursor import CursorFrame
from win2xcur.parser import ANIParser, CURParser


def to_cur(frame: CursorFrame) -> bytes:
    header = CURParser.ICON_DIR.pack(0, CURParser.ICO_TYPE_CUR, len(frame))
    directory: List[bytes] = []
    image_data: List[bytes] = []
    offset = CURParser.ICON_DIR.size + len(frame) * CURParser.ICON_DIR_ENTRY.size

    for image in frame:
        clone = image.image.clone()
        if clone.width > 256 or clone.height > 256:
            raise ValueError(f'Image too big for CUR format: {clone.width}x{clone.height}')
        blob = clone.make_blob('png')
        image_data.append(blob)
        x_offset, y_offset = image.hotspot
        directory.append(CURParser.ICON_DIR_ENTRY.pack(
            clone.height & 0xFF, clone.height & 0xFF, 0, 0, x_offset, y_offset, len(blob), offset
        ))
        offset += len(blob)

    return b''.join(chain([header], directory, image_data))


def get_ani_cur_list(frames: List[CursorFrame]) -> bytes:
    io = BytesIO()
    for frame in frames:
        cur_file = to_cur(frame)
        io.write(ANIParser.CHUNK_HEADER.pack(ANIParser.ICON_CHUNK, len(cur_file)))
        io.write(cur_file)
        if len(cur_file) & 1:
            io.write(b'\0')
    return io.getvalue()


def get_ani_rate_chunk(frames: List[CursorFrame]) -> bytes:
    io = BytesIO()
    io.write(ANIParser.CHUNK_HEADER.pack(ANIParser.RATE_CHUNK, ANIParser.UNSIGNED.size * len(frames)))
    for frame in frames:
        io.write(ANIParser.UNSIGNED.pack(int(round(frame.delay * 60))))
    return io.getvalue()


def to_ani(frames: List[CursorFrame]) -> bytes:
    ani_header = ANIParser.ANIH_HEADER.pack(
        ANIParser.ANIH_HEADER.size, len(frames), len(frames), 0, 0, 32, 1, 1, ANIParser.ICON_FLAG
    )

    cur_list = get_ani_cur_list(frames)
    chunks = [
        ANIParser.CHUNK_HEADER.pack(ANIParser.HEADER_CHUNK, len(ani_header)),
        ani_header,
        ANIParser.RIFF_HEADER.pack(ANIParser.LIST_CHUNK, len(cur_list) + 4, ANIParser.FRAME_TYPE),
        cur_list,
        get_ani_rate_chunk(frames),
    ]
    body = b''.join(chunks)
    riff_header: bytes = ANIParser.RIFF_HEADER.pack(ANIParser.SIGNATURE, len(body) + 4, ANIParser.ANI_TYPE)
    return riff_header + body


def to_smart(frames: List[CursorFrame]) -> Tuple[str, bytes]:
    if len(frames) == 1:
        return '.cur', to_cur(frames[0])
    else:
        return '.ani', to_ani(frames)



from typing import Any

import numpy as np


def premultiply_alpha(source: bytes) -> bytes:
    buffer: np.ndarray[Any, np.dtype[np.double]] = np.frombuffer(source, dtype=np.uint8).astype(np.double)
    alpha = buffer[3::4] / 255.0
    buffer[0::4] *= alpha
    buffer[1::4] *= alpha
    buffer[2::4] *= alpha
    return buffer.astype(np.uint8).tobytes()
